/****************************************************************************************
*  YAROCK                                                                               *
*  Copyright (c) 2010-2013 Sebastien amardeilh <sebastien.amardeilh+yarock@gmail.com>   *
*                                                                                       *
*  This program is free software; you can redistribute it and/or modify it under        *
*  the terms of the GNU General Public License as published by the Free Software        *
*  Foundation; either version 2 of the License, or (at your option) any later           *
*  version.                                                                             *
*                                                                                       *
*  This program is distributed in the hope that it will be useful, but WITHOUT ANY      *
*  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A      *
*  PARTICULAR PURPOSE. See the GNU General Public License for more details.             *
*                                                                                       *
*  You should have received a copy of the GNU General Public License along with         *
*  this program.  If not, see <http://www.gnu.org/licenses/>.                           *
*****************************************************************************************/


#include "mpris1.h"
#include "mpris2.h"
#include "mpris_common.h"

#include "EnginePlayer/engineplayer.h"
#include "Widget/audiocontrols.h"
#include "global_actions.h"
#include "utilities.h"
#include "constants.h"
#include "debug.h"

#include <QCoreApplication>
#include <QDBusConnection>
#include <QDBusMetaType>
#include <QtDBus>
/*
********************************************************************************
*                                                                              *
*    Mpris2                                                                    *
*                                                                              *
********************************************************************************
*/
Mpris2::Mpris2(Mpris1* mpris1, QObject* parent) : QObject(parent)
{
    //Debug::warning() << "##### MPRIS2 create";

    //MPRISv2.0
    new Mpris2Root(this);
    new Mpris2Player(mpris1, this);

    if (!QDBusConnection::sessionBus().registerService("org.mpris.MediaPlayer2.yarock")) {
      Debug::warning() << "Failed to register mpris2 dbus connection";
      return;
    }

    QDBusConnection::sessionBus().registerObject("/org/mpris/MediaPlayer2", this);
}


Mpris2::~Mpris2()
{
    QDBusConnection::sessionBus().unregisterService("org.mpris.MediaPlayer2.yarock");
}


/*
********************************************************************************
*                                                                              *
*    Mpris2Root                                                                *
*                                                                              *
********************************************************************************
*/
Mpris2Root::Mpris2Root(QObject *parent) : QDBusAbstractAdaptor(parent)
{
}

Mpris2Root::~Mpris2Root()
{

}

bool Mpris2Root::canQuit() const
{
    return true;
}

bool Mpris2Root::canRaise() const
{
    return false;
}

bool Mpris2Root::hasTrackList() const
{
    return false;
}

QString Mpris2Root::desktopEntry() const
{
    return QString("%1").arg(APP_NAME);
}


QString Mpris2Root::identity() const
{
    return QString("%1 %2").arg(APP_NAME,VERSION);
}

QStringList Mpris2Root::supportedMimeTypes() const
{
  static QStringList res = QStringList()
                           << "application/ogg"
                           << "application/x-ogg"
                           << "application/x-ogm-audio"
                           << "audio/aac"
                           << "audio/mp4"
                           << "audio/mpeg"
                           << "audio/mpegurl"
                           << "audio/ogg"
                           << "audio/vnd.rn-realaudio"
                           << "audio/vorbis"
                           << "audio/x-flac"
                           << "audio/x-mp3"
                           << "audio/x-mpeg"
                           << "audio/x-mpegurl"
                           << "audio/x-ms-wma"
                           << "audio/x-musepack"
                           << "audio/x-oggflac"
                           << "audio/x-pn-realaudio"
                           << "audio/x-scpls"
                           << "audio/x-speex"
                           << "audio/x-vorbis"
                           << "audio/x-vorbis+ogg"
                           << "audio/x-wav"
                           << "video/x-ms-asf"
                           << "x-content/audio-player";
  return res;
}

QStringList Mpris2Root::supportedUriSchemes() const
{
    static QStringList res = QStringList()
                           << "file"
                           << "http";
    return res;
}

void Mpris2Root::Quit()
{
    qApp->quit();
}

void Mpris2Root::Raise()
{
}

/*
********************************************************************************
*                                                                              *
*    Mpris2Player                                                              *
*                                                                              *
********************************************************************************
*/
Mpris2Player::Mpris2Player(Mpris1* mpris1, QObject *parent) : QDBusAbstractAdaptor(parent), m_mpris1(mpris1)
{
    QObject::connect(EnginePlayer::instance(), SIGNAL(engineStateChanged()), SLOT(slot_engineStateChanged()));
    QObject::connect(EnginePlayer::instance(), SIGNAL(playingInfoChanged()), SLOT(slot_currentSongChanged()));
}

Mpris2Player::~Mpris2Player() {}


QString Mpris2Player::PlaybackStatus() const
{
    switch (EnginePlayer::instance()->engineState()) {
      case ENGINE::PLAYING : return "Playing"; break;
      case ENGINE::PAUSED  : return "Paused";break;
      default : return "Stopped";break;
    }
}

QString Mpris2Player::LoopStatus() const
{
    //! class RepeatControl ==> enum RepeatMode {RepeatOff = 0,RepeatTrack = 1, RepeatAll = 2 };
    int state = RepeatControl::instance()->getState();

    if(state == 0)          /*RepeatOff */
      return QString("None");
    else if(state == 1)     /*RepeatTrack */
      return QString("Track");
    else if (state == 2)    /*RepeatAll */
      return QString("Playlist");
    else
      return QString("None");
}

void Mpris2Player::SetLoopStatus(const QString& value)
{
    //Debug::warning() << "##### MPRIS2 Mpris2Player SetLoopStatus";

    //! class RepeatControl ==> enum RepeatMode {RepeatOff = 0,RepeatTrack = 1, RepeatAll = 2 };
    int state = 0;

    if (value == "None")
      state = 0;
    else if (value == "Track")
      state = 1;
    else if (value == "Playlist")
      state = 2;

    RepeatControl::instance()->setState(state);
}

double Mpris2Player::Rate() const
{
    return 1.0;
}

void Mpris2Player::SetRate(double rate)
{
    if(rate == 0)
      ACTIONS()->value(ENGINE_PLAY)->trigger();
}

bool Mpris2Player::Shuffle() const
{
    return (ShuffleControl::instance()->getState() == 1) ? true : false;
}

void Mpris2Player::SetShuffle(bool value)
{
    ShuffleControl::instance()->setState(value);
}

QVariantMap Mpris2Player::Metadata() const
{
    return m_last_metadata;
}


double Mpris2Player::Volume() const
{
    return (EnginePlayer::instance()->volume() / 100);
}

void Mpris2Player::SetVolume(double value)
{
    EnginePlayer::instance()->setVolume(value * 100);
}


qlonglong Mpris2Player::Position() const
{
    /* currentTime() = in milliseconds */
    return EnginePlayer::instance()->mediaObject()->currentTime() / 1000;
}

double Mpris2Player::MaximumRate() const
{
    return 1.0;
}

double Mpris2Player::MinimumRate() const
{
    return 1.0;
}

bool Mpris2Player::CanGoNext() const
{
    return m_mpris1->player()->GetCaps() & CAN_GO_NEXT;
}

bool Mpris2Player::CanGoPrevious() const
{
    return m_mpris1->player()->GetCaps() & CAN_GO_PREV;
}

bool Mpris2Player::CanPlay() const
{
  return m_mpris1->player()->GetCaps() & CAN_PLAY;
}

// Taken from clementine ;
// This one's a bit different than MPRIS 1 - we want this to be true even when
// the song is already paused or stopped.
bool Mpris2Player::CanPause() const
{
    return m_mpris1->player()->GetCaps() & CAN_PAUSE
        || PlaybackStatus() == "Paused"
        || PlaybackStatus() == "Stopped";
}

bool Mpris2Player::CanSeek() const
{
    return m_mpris1->player()->GetCaps() & CAN_SEEK;
}

bool Mpris2Player::CanControl() const
{
    return true;
}

void Mpris2Player::Next()
{
    //Debug::warning() << "##### MPRIS2 Mpris2Player Next";
    if(CanGoNext())
      ACTIONS()->value(ENGINE_PLAY_NEXT)->trigger();
}

void Mpris2Player::Previous()
{
    //Debug::warning() << "##### MPRIS2 Mpris2Player Previous";
    if(CanGoPrevious())
      ACTIONS()->value(ENGINE_PLAY_PREV)->trigger();
}

void Mpris2Player::Pause()
{
    //Debug::warning() << "##### MPRIS2 Mpris2Player Pause";
    if(CanPause() && EnginePlayer::instance()->engineState() != ENGINE::PAUSED)
      ACTIONS()->value(ENGINE_PLAY)->trigger();
}

void Mpris2Player::PlayPause()
{
    //Debug::warning() << "##### MPRIS2 Mpris2Player PlayPause";
    if (CanPause())
      ACTIONS()->value(ENGINE_PLAY)->trigger();
}

void Mpris2Player::Stop()
{
    //Debug::warning() << "##### MPRIS2 Mpris2Player Stop";
    ACTIONS()->value(ENGINE_STOP)->trigger();
}

void Mpris2Player::Play()
{
    //Debug::warning() << "##### MPRIS2 Mpris2Player Play";
    if(CanPlay())
      ACTIONS()->value(ENGINE_PLAY)->trigger();
}

void Mpris2Player::Seek(qlonglong offset)
{
    //Debug::warning() << "##### MPRIS2 Mpris2Player Seek";
    if(CanSeek())
    {
      int msec = EnginePlayer::instance()->mediaObject()->currentTime()
                + offset*1000;
      EnginePlayer::instance()->seek(msec);

      Debug::debug() << "MPRIS2 : seek msec : " << msec;
    }
}

void Mpris2Player::SetPosition(const QDBusObjectPath& trackId, qlonglong offset)
{
Q_UNUSED(trackId)
Q_UNUSED(offset)
/* TODO */
}

void Mpris2Player::OpenUri(const QString& uri)
{
Q_UNUSED(uri)
/* TODO */
}

/*
--------------------------------------------------------------------------------
                Private slots
--------------------------------------------------------------------------------
*/
void Mpris2Player::slot_engineStateChanged()
{
    //Debug::debug() << "##### MPRIS2 : slot_engineStateChanged";
    ENGINE::E_ENGINE_STATE engine_state = EnginePlayer::instance()->engineState();

    if(engine_state != ENGINE::PLAYING && engine_state != ENGINE::PAUSED) {
      // clear metadata
      m_last_metadata= QVariantMap();
      EmitNotification("Metadata");
    }
    EmitNotification("PlaybackStatus");
}

void Mpris2Player::slot_currentSongChanged()
{
    //Debug::debug() << "##### MPRIS2 : slot_currentSongChanged";
    // fill metadata
    m_last_metadata = GetMetadata(EnginePlayer::instance()->playingTrack());

    EmitNotification("Metadata");
}

QVariantMap Mpris2Player::GetMetadata(const MEDIA::TrackPtr track)
{
    //Debug::debug() << "##### MPRIS2 : GetMetadata";

    QVariantMap ret;
    if(track->type() == TYPE_TRACK)
    {
      AddMetadata("xesam:url",         track->url,                &ret);
      AddMetadata("xesam:artist",      track->artist,             &ret);
      AddMetadata("xesam:title",       track->title,              &ret);
      AddMetadata("xesam:album",       track->album,              &ret);
      AddMetadata("mpris:length",      track->duration,           &ret);  // The duration in seconds
      AddMetadata("xesam:trackNumber", QString::number(track->num), &ret);
      AddMetadata("xesam:genre",       track->genre, &ret);

      const QString coverpath = UTIL::CONFIGDIR + "/albums/" + track->coverName();
      if( QFile(coverpath).exists() ) {
        //Debug::debug() << "MPRIS2 : slot_currentSongChanged add mpris::artUrl: " << coverpath;
        AddMetadata("mpris:artUrl",      coverpath,  &ret);
      }

      if (track->rating != -1.0)
       AddMetadata("rating", track->rating * 5, &ret);
    }
    else  if(track->type() == TYPE_STREAM)
    {
     //! TODO
    }

    return ret;
}


void Mpris2Player::EmitNotification(const QString& name)
{
    //Debug::debug() << "#### MPRIS2 : EmitNotification";

    QVariant value;
    if      (name == "PlaybackStatus")  value = PlaybackStatus();
    //else if (name == "LoopStatus")      value = LoopStatus();
    else if (name == "Shuffle")         value = Shuffle();
    else if (name == "Metadata")        value = Metadata();
    else if (name == "Volume")          value = Volume();
    else if (name == "Position")        value = Position();

    if (value.isValid())
    {
      QDBusMessage msg = QDBusMessage::createSignal(
        "/org/mpris/MediaPlayer2", "org.freedesktop.DBus.Properties", "PropertiesChanged");

      QVariantMap map;
      map.insert(name, value);
      QVariantList args = QVariantList()
                      << "org.mpris.MediaPlayer2.Player"
                      << map
                      << QStringList();
     msg.setArguments(args);
     QDBusConnection::sessionBus().send(msg);
    }
}
