/****************************************************************************************
*  YAROCK                                                                               *
*  Copyright (c) 2010-2013 Sebastien amardeilh <sebastien.amardeilh+yarock@gmail.com>   *
*                                                                                       *
*  This program is free software; you can redistribute it and/or modify it under        *
*  the terms of the GNU General Public License as published by the Free Software        *
*  Foundation; either version 2 of the License, or (at your option) any later           *
*  version.                                                                             *
*                                                                                       *
*  This program is distributed in the hope that it will be useful, but WITHOUT ANY      *
*  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A      *
*  PARTICULAR PURPOSE. See the GNU General Public License for more details.             *
*                                                                                       *
*  You should have received a copy of the GNU General Public License along with         *
*  this program.  If not, see <http://www.gnu.org/licenses/>.                           *
*****************************************************************************************/

//! local
#include "scrobblerlastfm.h"
#include "constants.h"
#include "utilities.h"
#include "debug.h"

#include "networkaccess.h"


//! Qt
#include <QNetworkAccessManager>
#include <QNetworkRequest>
#include <QNetworkReply>

#include <QtXml>

#include <QCryptographicHash>
#include <QFile>
#include <QTimer>
#include <QDir>
#include <QTextStream>
#include <QDateTime>

#include <QDesktopServices>


/* YAROCK API KEY
  API Key is b3717637c18071e1619e92ee2c3eb0f8
  secret is 8192599d44d34b27c520d597d34f3714
*/

/*******************************************************************************
    LastFM namespace
*******************************************************************************/
namespace LastFm {

  namespace Global {
        QString api_key;
        QString session_key;
        QString secret_key;
        QString auth_token;
        QString username;
  }

  inline QString md5( const QByteArray& src )
  {
    QByteArray const digest = QCryptographicHash::hash( src, QCryptographicHash::Md5 );
    return QString::fromLatin1( digest.toHex() ).rightJustified( 32, '0' ).toLower();
  }

  // Returns signature of given request
  QString sign(QUrl request)
  {
    // sort the list of query items alphabetically
    QList<QPair<QString, QString> > list = request.queryItems();
    qSort(list);

    QUrl url;
    url.setQueryItems(list);
    // convert name1=value1&name2=value2... to name1value1name2value2
    QString str = url.toString();
    str.remove(url.queryPairDelimiter());
    str.remove(url.queryValueDelimiter());
    str.remove(0,1); // remove trailing "?"
    str.append(LastFm::Global::secret_key);
    QByteArray ba;
    ba.append(str);

    return QString(QCryptographicHash::hash(ba, QCryptographicHash::Md5).toHex());
  }

}

/*
********************************************************************************
*                                                                              *
*    Class LastFm::Scrobbler                                                   *
*                                                                              *
********************************************************************************
*/
LastFm::Scrobbler::Scrobbler()
{
    Debug::debug() << "--- LastFm : Create Scrobbler";

    m_currentTrack    = 0;
    m_cache           = new LastFm::Cache(this);

    //! load settings
    QSettings settings(UTIL::CONFIGFILE,QSettings::IniFormat,this);
    settings.beginGroup("Scrobbler");

    LastFm::Global::username      = settings.value("username", QString()).toString();
    LastFm::Global::session_key   = settings.value("key", QString()).toString();
    LastFm::Global::api_key       = "b3717637c18071e1619e92ee2c3eb0f8";
    LastFm::Global::secret_key    = "8192599d44d34b27c520d597d34f3714";
    settings.endGroup();
}


LastFm::Scrobbler::~Scrobbler()
{
    //Debug::debug() << "--- Scrobbler ->delete";

    if(isAuthenticated())
      scrobble();

    //! save settings
    QSettings settings(UTIL::CONFIGFILE,QSettings::IniFormat,this);
    settings.beginGroup("Scrobbler");
    settings.setValue("username", LastFm::Global::username);
    settings.setValue("key", LastFm::Global::session_key);
    settings.endGroup();
    settings.sync();

    //! delete
    //Debug::debug() << "--- Scrobbler ->delete cache";
    delete m_cache;
}

bool LastFm::Scrobbler::isAuthenticated() const
{
  return !LastFm::Global::session_key.isEmpty();
}


void LastFm::Scrobbler::signIn(const QString& username, const QString& password)
{
    //Debug::debug() << "--- LastFm : Sign In request...username" << username << " password " << password;
    QNetworkRequest request;
    request.setHeader(QNetworkRequest::ContentTypeHeader,"application/x-www-form-urlencoded");

    QUrl url;
    url.setUrl("http://ws.audioscrobbler.com/2.0/");
    url.addQueryItem("api_key", LastFm::Global::api_key);
    url.addQueryItem("method", "auth.getMobileSession");
    url.addQueryItem("username",username);
    url.addQueryItem("authToken", LastFm::md5((username + LastFm::md5(password.toUtf8())).toUtf8()));
    url.addQueryItem("api_sig", LastFm::sign(url));
    request.setUrl(url.toString());

    LastFm::Global::auth_token = LastFm::md5( (username + LastFm::md5(password.toUtf8())).toUtf8() );

    /* post request */
    NetworkReply *reply = HTTP()->get(url);
    connect(reply, SIGNAL(finished(QNetworkReply*)), SLOT(slotSignInFinished(QNetworkReply*)));
}


void LastFm::Scrobbler::slotSignInFinished(QNetworkReply *reply)
{
    int status = reply->attribute(QNetworkRequest::HttpStatusCodeAttribute).toInt();
    if (status != 200)
    {
        Debug::debug() << "--- LastFm : Sign In request failed, http error" << status;
        raiseError(-1);
        return;
    }

    // Read the XML document
    QDomDocument document;
    document.setContent(reply->readAll());


    QDomElement lfm = document.documentElement();
    if (lfm.attribute("status", "") == "ok")
    {
        QDomElement name            = lfm.elementsByTagName("name").at(0).toElement();
        QDomElement key             = lfm.elementsByTagName("key").at(0).toElement();
        LastFm::Global::session_key = key.childNodes().at(0).nodeValue();
        LastFm::Global::username    = name.childNodes().at(0).nodeValue();

        Debug::debug() << "--- LastFm : Sign in Successfull";
        Debug::debug() << "--- LastFm : Sign in Successfull LastFm::Global::session_key " << LastFm::Global::session_key;
        Debug::debug() << "--- LastFm : Sign in Successfull LastFm::Global::username " << LastFm::Global::username;
    }

    emit signInFinished();
}

void LastFm::Scrobbler::signOut()
{
    LastFm::Global::username.clear();
    LastFm::Global::session_key.clear();
}


void LastFm::Scrobbler::raiseError(int code)
{
    QString errmsg;

    switch (code)
    {
    case -1: errmsg = tr("Invalid server response");                   break;
    case  2: errmsg = "Invalid service - This service does not exist"; break;
    case  3: errmsg = "Invalid Method - No method with that name in this package"; break;
    case  4: errmsg = "Authentication Failed - You do not have permissions to access the service"; break;
    case  5: errmsg = "Invalid format - This service doesn't exist in that format"; break;
    case  6: errmsg = "Invalid parameters - Your request is missing a required parameter";break;
    case  7: errmsg = "Invalid resource specified"; break;
    case  9: errmsg = "Invalid session key - Please re-authenticate"; break;
    case 10: errmsg = "Invalid API key - You must be granted a valid key by last.fm"; break;
    case 11: // Service Offline - This service is temporarily offline. Try again later.
    case 16: // There was a temporary error processing your request. Please try again
        QTimer::singleShot(1200000, m_cache, SLOT(submit()));
        return;
        break;
    case 13: errmsg = "Invalid method signature supplied"; break;
    case 14: errmsg = tr("Authentication Failed - This token has not been authorized"); break;
    case 26: errmsg = "Suspended API key - Access for your account has been suspended, please contact Last.fm"; break;
    }

    emit error(errmsg, code);
}


void LastFm::Scrobbler::scrobble()
{
    if (m_currentTrack)
        m_currentTrack->scrobble();
    else
        m_cache->submit();
}

void LastFm::Scrobbler::scrobble(LastFm::Track *track)
{
    track->scrobble();
}

void LastFm::Scrobbler::nowPlaying(LastFm::Track *track)
{
    m_currentTrack = track;
    if (track)
        track->nowPlaying();
}



/*
********************************************************************************
*                                                                              *
*    Class LastFm::Track                                                       *
*                                                                              *
********************************************************************************
*/
LastFm::Track::Track(LastFm::Scrobbler *scrobbler, QString artist, QString trackTitle,
                     QString album, int trackLength, QString genre, int trackNumber,
                     uint playbackStart):
        m_scrobbler(scrobbler),
        m_artist(artist),
        m_trackTitle(trackTitle),
        m_album(album),
        m_trackLength(trackLength),
        m_genre(genre),
        m_trackNumber(trackNumber),
        m_playbackStart(playbackStart)
{
}


void LastFm::Track::scrobble()
{
    m_playbackLength  += QDateTime::currentDateTime().toTime_t() - m_unpauseTime;
    Debug::debug() << "--- LastFm scrobble track m_playbackLength" << m_playbackLength;
    Debug::debug() << "--- LastFm scrobble track m_trackLength" << m_trackLength;

    // The track must be played for at least half of it's duration or at least 4 minutes, whatever
    // occurs first and must be longer then 30 seconds to be scrobbled
    if (((m_playbackLength < m_trackLength / 2) && (m_playbackLength < 240)) || (m_trackLength < 30)) {
        Debug::debug() << "--- LastFm : Track" << m_trackTitle << "was not played long enough or is too short, not scrobbling";
        return;
    }

    // Unset the current track - when we are scrobbling, this track is no more current
    m_scrobbler->setCurrentTrack(0);

    // Add itself to the cache
    if (!m_scrobbler->cache())
        return;

    m_scrobbler->cache()->add(this);
    m_scrobbler->cache()->submit();
}


void LastFm::Track::nowPlaying()
{
    QUrl requestUrl("http://ws.audioscrobbler.com/2.0/");
    requestUrl.addQueryItem("method", "track.updateNowPlaying");
    requestUrl.addQueryItem("track", m_trackTitle);

    QByteArray data;
    QUrl params;
    params.addQueryItem("album", m_album);
    params.addQueryItem("api_key", LastFm::Global::api_key);
    params.addQueryItem("artist", m_artist);
    params.addQueryItem("duration", QString::number(m_trackLength));
    params.addQueryItem("method", "track.updateNowPlaying");
    params.addQueryItem("sk", LastFm::Global::session_key);
    params.addQueryItem("token", LastFm::Global::auth_token);
    params.addQueryItem("track", m_trackTitle);
    params.addQueryItem("trackNumber", QString::number(m_trackNumber));
    params.addQueryItem("api_sig", LastFm::sign(params));
    // Remove the trailing "?" from the params, since we are doing POST
    data.append(params.toString().remove(0,1));

    /* post request */
    NetworkReply *reply = HTTP()->post(requestUrl, data);
    connect(reply, SIGNAL(finished(QNetworkReply*)), SLOT(scrobbled(QNetworkReply*)));

    // Set the track as current
    m_scrobbler->setCurrentTrack(this);

    // We started playing
    m_playbackLength = 0;
    m_unpauseTime = m_playbackStart;
}


void LastFm::Track::love()
{
    QUrl requestUrl("http://ws.audioscrobbler.com/2.0/");
    requestUrl.addQueryItem("method", "track.love");
    requestUrl.addQueryItem("track", m_trackTitle);

    QByteArray data;
    QUrl params;
    params.addQueryItem("api_key", LastFm::Global::api_key);
    params.addQueryItem("artist", m_artist);
    params.addQueryItem("method", "track.love");
    params.addQueryItem("sk", LastFm::Global::session_key);
    params.addQueryItem("token", LastFm::Global::auth_token);
    params.addQueryItem("track", m_trackTitle);
    params.addQueryItem("api_sig", LastFm::sign(params));
    data.append(params.toString().remove(0,1));

    
    /* post request */
    NetworkReply *reply = HTTP()->post(requestUrl, data);
    connect(reply, SIGNAL(finished(QNetworkReply*)), SLOT(scrobbled(QNetworkReply*)));
}


void LastFm::Track::pause(bool pause)
{
    if (pause) {
        m_playbackLength += QDateTime::currentDateTime().toTime_t() - m_unpauseTime;
    } else {
        m_unpauseTime = QDateTime::currentDateTime().toTime_t();
    }
}


void LastFm::Track::scrobbled(QNetworkReply *reply)
{
    QDomDocument document;
    document.setContent(reply->readAll());


    QString method = reply->request().url().queryItemValue("method");

    QDomElement lfm = document.documentElement();
    if (lfm.attribute("status", "") == "ok") {
        Debug::debug() << "--- LastFm" << method << "for" << reply->request().url().queryItemValue("track") << "successfull";
        if (method == "track.love")
            emit loved();
        else
            emit scrobbled();

    } else {
        Debug::debug() << "--- LastFm" << method << "for" << reply->request().url().queryItemValue("track") << "failed:";
        QDomElement err = lfm.childNodes().at(0).toElement();
        Debug::debug() << err.childNodes().at(0).nodeValue();

        m_scrobbler->raiseError(err.attribute("code", 0).toInt());
    }
}


/*
********************************************************************************
*                                                                              *
*    Class LastFm::Cache                                                       *
*                                                                              *
********************************************************************************
*/
LastFm::Cache::Cache(LastFm::Scrobbler *scrobbler):
        m_scrobbler(scrobbler),
        m_resubmitTimer(0)
{
    loadCache();
}


LastFm::Cache::~Cache()
{
    // Submit entire cache
    submit();

    // Save what's left
    QFile file(UTIL::CONFIGFILE + QDir::separator() + "lastfmcache.xml");
    if (!file.open(QIODevice::WriteOnly))
        return;

    QDomDocument document;

    QDomElement rootElement = document.createElement("tracks");
    document.appendChild(rootElement);

    for (int i = 0; i < m_cache.size(); i++)
    {
        QDomElement trackElement = document.createElement("track");

        LastFm::Track *track = m_cache.at(i);
        trackElement.setAttribute("track", track->trackTitle());
        trackElement.setAttribute("artist", track->artist());
        trackElement.setAttribute("album", track->album());
        trackElement.setAttribute("genre", track->genre());
        trackElement.setAttribute("length", track->trackLength());
        trackElement.setAttribute("trackNumber", track->trackNumber());
        trackElement.setAttribute("playbackStart", track->playbackStart());

        rootElement.appendChild(trackElement);
    }

    // Save the stream
    QTextStream ts(&file);
    ts << document.toString();

    // Delete all tracks in cache
    qDeleteAll(m_cache);
}


void LastFm::Cache::add(LastFm::Track *track)
{
    Debug::debug() << "--- LastFm : Adding track" << track->trackTitle() << "to cache";
    if (track)
        m_cache.append(track);
}


void LastFm::Cache::submit()
{
    QUrl requestUrl("http://ws.audioscrobbler.com/2.0/");
    requestUrl.addQueryItem("method", "track.scrobble");

    /* set post param */
    QByteArray data;
    QUrl params;

    Debug::debug() << "--- LastFm : Scrobbling " << m_cache.size() << " tracks";

    params.addQueryItem("api_key", LastFm::Global::api_key);
    params.addQueryItem("method", "track.scrobble");
    params.addQueryItem("sk", LastFm::Global::session_key);
    params.addQueryItem("token", LastFm::Global::auth_token);
    for (int i = 0; i < m_cache.size(); i++)
    {
        if (m_cache.at(i)) {
            params.addQueryItem("album["+QString::number(i)+"]", m_cache.at(i)->album());
            params.addQueryItem("artist["+QString::number(i)+"]", m_cache.at(i)->artist());
            params.addQueryItem("duration["+QString::number(i)+"]", QString::number(m_cache.at(i)->trackLength()));
            params.addQueryItem("timestamp["+QString::number(i)+"]", QString::number(m_cache.at(i)->playbackStart()));
            params.addQueryItem("track["+QString::number(i)+"]", m_cache.at(i)->trackTitle());
            params.addQueryItem("trackNumber["+QString::number(i)+"]", QString::number(m_cache.at(i)->trackNumber()));
        }
    }
    params.addQueryItem("api_sig", LastFm::sign(params));

    data.append(params.toString().remove(0,1));

    /* post request */
    NetworkReply *reply = HTTP()->post(requestUrl, data);
    connect(reply, SIGNAL(finished(QNetworkReply*)), SLOT(submitted(QNetworkReply*)));
}


void LastFm::Cache::submitted(QNetworkReply *reply)
{
    QDomDocument document;
    document.setContent(reply->readAll());

    QString method = reply->request().url().queryItemValue("method");

    QDomElement lfm = document.documentElement();
    if (lfm.attribute("status", "") == "ok") {
        Debug::debug() << "--- LastFm : " << method << "for tracks in cache successfull, purging cache";
        qDeleteAll(m_cache);
        m_cache.clear();

        // If we were able to submit track BEFORE the timer or this is timer's event, then
        // destroy the timer so that it won't tick anymore
        if (m_resubmitTimer)
            delete m_resubmitTimer;

    } else {
        Debug::debug() << "--- LastFm : " << method << "for  cache failed:";
        QDomElement err = lfm.childNodes().at(0).toElement();
        Debug::debug() << err.childNodes().at(0).nodeValue();

        m_scrobbler->raiseError(err.attribute("code", 0).toInt());

        m_resubmitTimer = new QTimer();
        // let's try again in 2 minutes
        m_resubmitTimer->setInterval(120000);
        connect(m_resubmitTimer, SIGNAL(timeout()),this, SLOT(submit()));
    }
}


void LastFm::Cache::loadCache()
{
    m_cache.clear();

    QFile file(UTIL::CONFIGFILE + QDir::separator() + "lastfmcache.xml");
    if (!file.open(QIODevice::ReadOnly))
        return;

    QDomDocument document;
    document.setContent(&file);

    QDomElement rootElement = document.documentElement();
    LastFm::Track *track;

    // Is the root element <tracks> ?
    if (rootElement.tagName() != "tracks")
        return;

    QDomNode trackNode = rootElement.firstChild();
    while ( !trackNode.isNull() )
    {
        QDomElement trackEl = trackNode.toElement();

        // Is the element valid?
        if (trackEl.isNull()) {
            trackNode = trackNode.nextSibling();
            continue;
        }

        // Is it a <track> element?
        if (trackEl.tagName() != "track") {
            trackNode = trackNode.nextSibling();
            continue;
        }

        track = new LastFm::Track(m_scrobbler,
                                  trackEl.attribute("artist", ""),
                                  trackEl.attribute("track", ""),
                                  trackEl.attribute("album", ""),
                                  trackEl.attribute("length", 0).toInt(),
                                  trackEl.attribute("genre", ""),
                                  trackEl.attribute("trackNumber", 0).toInt(),
                                  trackEl.attribute("playbackStart", 0).toInt());
        m_cache.append(track);

        trackNode = trackNode.nextSibling();
    }
}
